using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using Newtonsoft.Json;
using REST_JSON_API;
using UnityEngine;
using UnityEngine.Events;
using UnityEngine.Networking;

#nullable enable annotations

/// <summary>
/// This class keeps track of the currently active scroll and communicates with the FrameIT Server.
/// It represents the model of the Model-View-Controller pattern.
/// </summary>
[Serializable]
public class ActiveScroll : MonoBehaviour
{

    /** <summary>Event that is invoked when a scroll button is clicked, e.g. magic button. You can use it to handle custom buttons.</summary> */
    public UnityEvent<ScrollButton> OnButtonClick = new();
    /** <summary>Notifies about available hints.</summary> */
    public UnityEvent<IReadOnlyList<string>> HintAvailableEvent = new();
    /** <summary>Event that is invoked when a fact has been assigned to a slot and the ui needs to update.</summary> */
    public UnityEvent<string,SlotAssignment> OnFactAssignmentUpdated = new();
    /** <summary>Event that is invoked when the MMT server returned an error.</summary> */
    public UnityEvent<string> OnMMTServerCommunicationError = new();
    /** <summary>Event that is invoked when the scroll dynamic info returned errors in the current scroll assignment.</summary> */
    public UnityEvent<ScrollApplicationCheckingError[]> OnScrollDynamicInfoError = new();
    /** <summary>Event that is invoked when the scroll application returned errors in the current assignment.</summary> */
    public UnityEvent<ScrollApplicationCheckingError[]> OnScrollApplicationError = new();
    /** <summary>Event that is invoked when the OnMMTServerError or OnScrollApplicationFailed error state has been resolved and e.g. the error popup can be closed.</summary> */
    public UnityEvent OnCancelErrorState = new();
    /** <summary>Event that is invoked when the scroll has been solved.</summary> */
    public UnityEvent<IReadOnlyList<Fact>> OnScrollSolved = new();
    /** <summary>Event that is invoked when an other scroll has been selected.</summary> */
    public UnityEvent<Scroll> OnScrollChanged = new();
    /** <summary>Event that is invoked when the dynamic info of the scroll has been updated e.g. because a fact has been assigned.</summary> */
    public UnityEvent<Scroll> OnScrollDynamicInfoUpdated = new();


    public bool DynamicScrollDescriptionsActive = true;
    public bool AutomaticHintGenerationActive = true;

    private readonly IReadOnlyList<string> NoDynamicScroll = new List<string>()
    {
        // Insert ScrollURIS that are poorly optimized
    };


    /// <summary>
    /// Variable used to block that only one SendView is running at a time.
    /// </summary>
    private bool SendingViewDone = true;
    /// <summary>
    /// this is true while a fact assignment is enqueued
    /// </summary>
    private bool DynamicScrollInQue = false;
    /// <summary>
    /// this is true while a scroll application via magic button is enqueued
    /// </summary>
    private bool MagicInQue = false;
    private string currentMmtAnswer;

    /** <summary>List of available hints for the current partial assignment.</summary> */
    private List<Fact> LatestRenderedHints;
    /** <summary>List of available auto completions for the current partial assignment. Used to generate hints.</summary> */
    private List<ScrollAssignment> LatestBackwartsCompletions;

    private Scroll _Scroll;
    /// <summary>
    /// The currently active scroll
    /// </summary>
    public Scroll Scroll
    {
        get => _Scroll;
        set {
            _Scroll = value;
            
            // initialize the assignments dictionary with the references to the required fact slots as keys
            // and an empty SlotAssignment
            _Assignments = value.requiredFacts
                .ToDictionary(fact => fact.@ref.uri, _ => new SlotAssignment());
            
            // apply fact assignments that are pre defined by the stage (defaults or constrains), if there are any
            if (StageStatic.stage.solution.ScrollOverwrites
                    .TryGetValue(value.ScrollReference, out (string, int, bool)[] population)
            && population.Length > 0)
            {
                foreach ((string Id, int index, bool show) in population)
                {
                    var key = value.requiredFacts[index].@ref.uri;
                    _Assignments[key].fact = FactRecorder.AllFacts[Id];
                    _Assignments[key].IsVisible = show;
                }
            }
            // notify the observers that the scroll has changed
            OnScrollChanged.Invoke(value);
        }
    }
    /// <summary>
    /// The currently active scroll with rendered description and backward_completions (hints)
    /// </summary>
    public Scroll RenderedScroll { get; private set; }

    public class SlotAssignment
    {
        public Fact? fact = null;
        public bool IsVisible = true;
        public bool IsSet => fact is not null;
    }

    /// <summary>
    /// Maps the required facts (keys) to the assigned facts (values)
    /// </summary>
    private Dictionary<string, SlotAssignment> _Assignments = new();

    /// <summary>
    /// Maps the required facts (keys) to the assigned facts (values).
    /// </summary>
    public IReadOnlyDictionary<string, SlotAssignment> Assignments => _Assignments;

    /// <summary>
    /// The ScrollView should call this method when a button is clicked
    /// </summary>
    public void ButtonClicked(ScrollButton button)
    {
        switch (button)
        {
            // Magic button to apply the scroll
            case ApplyScrollButton:
                if ((Scroll is not null) && ! MagicInQue) StartCoroutine(SendScrollApplication());
                break;
            case HintScrollButton:
                HintButtonPressed((HintScrollButton)button);
                break;
            // add cases for future additional buttons here
        }
        // notify observers that a button was clicked
        OnButtonClick.Invoke(button);
    }

    /// <summary>
    /// Assigns the given fact to the given slotUri and calls the OnFactAssignmentUpdated Event
    /// </summary>
    /// <param name="slotUri"></param>
    /// <param name="fact"></param>
    /// <exception cref="ArgumentException"></exception>
    public void AssignFact(string slotUri, Fact fact)
    {
        if (Assignments.ContainsKey(slotUri))
        {
            var slot = _Assignments[slotUri];

            // return if the slot is already assigned to the given value (avoid unnecessary ui updates and FrameIT Server requests)
            //if ((fact is null && !slot.IsSet) || (slot.IsSet && slot.fact.Id == fact.Id)) return;
            if (fact is null && !slot.IsSet) return;

            slot.fact = fact;
            SendCurrentFactAssignments();
            OnFactAssignmentUpdated.Invoke(slotUri, slot);
        } else {
            throw new ArgumentException($"Can't assign fact, the given slot doesn't exist on the current scroll: '{slotUri}'");
        }
    }

    /// <summary>
    /// If the fact is not already assigned, assign it to the next empty slot, otherwise unassign it.
    /// And calls the OnFactAssignmentUpdated Event.
    /// </summary>
    /// <param name="fact"></param>
    /// <returns></returns>
    public string? AssignOrRemoveFactToNextEmptySlot(Fact fact)
    {
        // unassign the fact if it is already assigned and return
        try {
            var assignedSlot = _Assignments.First(pair => pair.Value.fact == fact).Key;
            AssignFact(assignedSlot, null);
            return null;
        } catch (Exception) {
            // the fact has not been assigned yet => continue
        }

        // find the first empty slot and assign the fact to it
        try {
            KeyValuePair<string, SlotAssignment> emptySlot = _Assignments.First(pair => !pair.Value.IsSet);
            AssignFact(emptySlot.Key, fact);
            return emptySlot.Key;
        } catch (Exception) {
            return null;
        }
    }


    #region MagicButton
    private IEnumerator SendScrollApplication()
    {
        // there is already a MagicButton coroutine waiting => exit
        if (MagicInQue) yield break;

        MagicInQue = true;
        // wait if SendView is not yet done or dynamic scroll is in the queue
        while (!SendingViewDone || DynamicScrollInQue)
            yield return null;
        MagicInQue = false;

        yield return SendView("/scroll/apply");

        if (currentMmtAnswer == null)
        {
            // OnMMTServerCommunicationError is already invoked in SendView
            Debug.LogError("Magic FAILED");
            yield break;
        }
        else
        {
            if (CommunicationEvents.VerboseURI)
                Debug.Log("Magic answers:\n" + currentMmtAnswer);

            System.DateTime deserializeTime = System.DateTime.UtcNow;
            ScrollApplicationInfo scrollApplication = JsonConvert.DeserializeObject<ScrollApplicationInfo>(currentMmtAnswer);
            Debug.LogFormat($"Answer deserialized in : {(System.DateTime.UtcNow - deserializeTime).TotalMilliseconds}ms");

            if (scrollApplication.acquiredFacts.Count == 0
                || scrollApplication.errors.Length > 0)
            {
                OnScrollApplicationError.Invoke(scrollApplication.errors);
            }
            else
            {
                OnCancelErrorState.Invoke();
            }

            var new_facts = __GenerateUnityFactsForAcquiredMMTFacts(scrollApplication.acquiredFacts);
            if (scrollApplication.acquiredFacts.Count >= 0 || scrollApplication.errors.Length == 0)
            {
                OnScrollSolved.Invoke(new_facts);
            }
        }
    }

    private List<Fact> __GenerateUnityFactsForAcquiredMMTFacts(List<MMTFact> pushoutFacts)
    {
        List<Fact> new_facts = new();
        Dictionary<string, string> old_to_new = new();
        System.DateTime parseTime = System.DateTime.UtcNow;

        bool sameStep = false;
        for (int i = 0; i < pushoutFacts.Count; i++)
        {
            List<Fact> new_list = Fact.MMTFactory(pushoutFacts[i].MapURIs(old_to_new));

            if (new_list.Count == 0)
            {
                Debug.LogWarning("Parsing on pushout-fact returned empty List -> One of the dependent facts does not exist or parsing failed");
                continue;
            }

            foreach (Fact new_fact in new_list)
            {
                Fact added = FactAdder.AddFactIfNotFound(new_fact, out bool exists, sameStep, null, Scroll.label);
                if (!exists)
                {
                    new_facts.Add(added);
                    CommunicationEvents.AnimateExistingFactEvent.Invoke(added.Id, FactWrapper.FactMaterials.Solution);
                    sameStep = true;
                }
                else
                {
                    // AnimateExistingFactEvent.Invoke(_new, FactWrapper.FactMaterials.Hint); // Automatically done in FactRecorder
                    old_to_new.Add(new_fact.Id, added.Id);
                }
            }

            //yield return null;
        }

        Debug.Log($"Facts parsed within {(System.DateTime.UtcNow - parseTime).TotalMilliseconds}ms");
        return new_facts;
    }

    #endregion MagicButton

    #region HintButton
    
    private void HintButtonPressed(HintScrollButton hintButton)
    {
        if (FactRecorder.AllFacts.ContainsKey(hintButton.SlotUri))
        {
            CommunicationEvents.AnimateExistingFactEvent.Invoke(
                    hintButton.SlotUri,
                    FactWrapper.FactMaterials.Hint
                );
        }

        Fact hintFact = LatestRenderedHints.Find(x => x.Id == hintButton.SlotUri); // "Dictionary"

        ScrollAssignment suitableCompletion =
            LatestBackwartsCompletions.Find((ScrollAssignment x) => x.fact.uri == hintButton.SlotUri); // "Dictionary"

        if (suitableCompletion != null && suitableCompletion.assignment is OMS assignment)
        {
            if (FactRecorder.AllFacts.ContainsKey(assignment.uri))
            {
                CommunicationEvents.AnimateExistingFactEvent.Invoke(
                    assignment.uri,
                    FactWrapper.FactMaterials.Hint
                );
            }
        }
        else if (hintFact != null)
        {
            if (FactRecorder.FindEquivalent(StageStatic.stage.factState.MyFactSpace, hintFact, out string found_key, out Fact _, out bool _, false))
                // existing fact -> Animate that
                CommunicationEvents.AnimateExistingFactEvent.Invoke(
                    found_key,
                    FactWrapper.FactMaterials.Hint
                );

            else
            {   // Generate new FactRepresentation and animate it
                CommunicationEvents.AnimateNonExistingFactEvent.Invoke(hintFact);
                CommunicationEvents.AnimateExistingFactEvent.Invoke(
                    hintButton.SlotUri,
                    FactWrapper.FactMaterials.Hint
                );
            }
        }
    }
    #endregion HintButton

    #region FactAssignment

    public void SendCurrentFactAssignments()
    {
        if (Scroll?.ScrollReference == null
         || NoDynamicScroll.Contains(Scroll.ScrollReference))
            return;

        if (AutomaticHintGenerationActive || DynamicScrollDescriptionsActive)
            StartCoroutine(_SendAssignments());
    }

    private IEnumerator _SendAssignments()
    {
        if (DynamicScrollInQue)
            yield break; // only need next in que to finish

        DynamicScrollInQue = true;
        while (!SendingViewDone)
            yield return null; // if we don't wait => server will crash
        DynamicScrollInQue = false;

        yield return SendView("/scroll/dynamic");

        if (currentMmtAnswer == null)
        {
            // OnMMTServerCommunicationError is already invoked in SendView
            Debug.LogError("Dynamic Scroll FAILED");
            yield break;
        }
        else
        {
            ScrollDynamicInfo scrollDynamicInfo;
            try
            {
                scrollDynamicInfo = JsonConvert.DeserializeObject<ScrollDynamicInfo>(currentMmtAnswer);
                //scrollDynamicInfo = IJSONsavable<ScrollDynamicInfo>.postprocess(scrollDynamicInfo); // DON'T! will remove hints
            }
            catch (JsonSerializationException ex)
            {
                Debug.LogException(ex);
                Debug.LogError("Could not Deserialize MMT answer for /scroll/dynamic\n" + currentMmtAnswer);
                OnMMTServerCommunicationError.Invoke("MMT Server communication error: could not deserialize message.");
                yield break;
            }

            ScrollApplicationCheckingError[] errors = scrollDynamicInfo.errors
                .Where(err => err.kind != "nonTotal") // expected
                .ToArray();

            if (errors.Length > 0) OnScrollDynamicInfoError.Invoke(errors);

            _ProcessScrollDynamicInfo(scrollDynamicInfo);
        }
    }

    private void _ProcessScrollDynamicInfo(ScrollDynamicInfo scrollDynamicInfo)
    {
        //TODO: more hints available in scrollDynamicInfo.rendered.requiredFacts

        LatestBackwartsCompletions = scrollDynamicInfo.backward_completions.Count > 0
            ? scrollDynamicInfo.backward_completions[0]
            : new List<ScrollAssignment>();

        List<string> hintUris = LatestBackwartsCompletions
            .Select(completion => completion.fact.uri)
            .ToList();

        //Update Scroll, process data for later hints and update Uri-List for which hints are available
        _processRenderedScroll(scrollDynamicInfo.rendered, hintUris);

        if (AutomaticHintGenerationActive)
            //Show that Hint is available for ScrollParameter
            HintAvailableEvent.Invoke(hintUris);

        return;

        void _processRenderedScroll(Scroll rendered, List<string> hintUris)
        {
            OnScrollDynamicInfoUpdated.Invoke(rendered);

            LatestRenderedHints = new();
            foreach (var requiredFact in rendered.requiredFacts)
            {
                //If ScrollFact is not assigned -> enable hint button
                if (Assignments.TryGetValue(requiredFact.@ref.uri, out SlotAssignment assignedFact) && ! assignedFact.IsSet)
                {
                    List<Fact> HintFactList = Fact.MMTFactory(requiredFact);

                    foreach (Fact HintFact in HintFactList)
                    {
                        hintUris.Add(HintFact.Id); // == rendered.requiredFacts[i].@ref.uri
                        LatestRenderedHints.Add(HintFact);
                    }
                }
            }

            return;
        }
    }


    #endregion FactAssignment

    /// <summary>
    /// Only one instance of the SendView coroutine should run at a time.
    /// </summary>
    private IEnumerator SendView(string endpoint)
    {
        SendingViewDone = false;

        //while (ParameterDisplays == null) // Wait for server
        //    yield return null;

        string body = prepareScrollAssignments();

        using UnityWebRequest www = UnityWebRequest.Put(CommunicationEvents.ServerAdress + endpoint, body);
        www.method = UnityWebRequest.kHttpVerbPOST;
        www.SetRequestHeader("Content-Type", "application/json");

        System.DateTime sendTime = System.DateTime.UtcNow;
        yield return www.SendWebRequest();
        //if (VerboseURI)
        Debug.LogFormat("Server answered in : {0}ms"
            , (System.DateTime.UtcNow - sendTime).TotalMilliseconds);

        if (www.result == UnityWebRequest.Result.ConnectionError
         || www.result == UnityWebRequest.Result.ProtocolError)
        {
            Debug.LogWarning(www.error);
            currentMmtAnswer = null;
            OnMMTServerCommunicationError.Invoke($"Server communication error: {www.error}");
            SendingViewDone = true;
            yield break;
        }
        else
        {
            while (!www.downloadHandler.isDone)
                yield return null;

            currentMmtAnswer = www.downloadHandler.text
                .Replace("\"float\":null", "\"float\":0.0"); // cannot convert null to value type
        }

        SendingViewDone = true;
        yield break;

        /// <summary>Maps the assigned slots to ScrollAssignment objects.
        /// Returns json serialized ScrollApplication containing the list of ScrollAssignments.</summary>
        string prepareScrollAssignments()
        {
            List<ScrollAssignment> assignmentList = Assignments.Where(assignment => assignment.Value.IsSet)
                .Select(assignment => new ScrollAssignment(assignment.Key, assignment.Value.fact.ServerDefinition))
                .ToList();

            return JsonConvert.SerializeObject(new ScrollApplication(Scroll.ScrollReference, assignmentList));
        }
    }
}

/// <summary>
/// Parent class for all scroll buttons, inherit from this class to implement custom buttons.
/// </summary>
public abstract class ScrollButton {}

public class ApplyScrollButton : ScrollButton {}
public class HintScrollButton : ScrollButton
{
    public readonly string SlotUri;
    public HintScrollButton(string slotUri) => SlotUri = slotUri;
}