using System;
using System.Linq;
using UnityEngine;

public class SphereGenerator : ShapeGenerator
{
    #region InspectorVariables
    [Header("Sphere values")]
    [SerializeField] protected float _radius = 1.0f;
    public float radius
    {
        get { return _radius; }
        set
        {
            _radius = value;
            GenerateShapeForAll();
        }
    }

    #endregion InspectorVariables

    #region Implementation
    protected override (Vector3[] vertices, int[] triangles) GenerateTopology()
        => CreateSphere(radius);

    /// <summary>
    /// Creates circle vertecies and triangles around the midPoint at (0,0,0)
    /// </summary>
    /// <param name="points"></param>
    /// <param name="invert"></param>
    /// <returns></returns>
public static (Vector3[], int[]) CreateSphere(float radius, Vector3 offset, bool invert = false)
{
    int pcount = 100; // Number of points around the circle at each latitude
    int steps = 20;  // Number of steps from the top to the bottom of the sphere (latitude steps)
    
    // Calculate the number of vertices and triangles
    int vertexCount = (pcount + 1) * (steps + 1);
    int triangleCount = pcount * steps * 2 * 3;

    Vector3[] vertices = new Vector3[vertexCount];
    int[] triangles = new int[triangleCount];

    int vertexIndex = 0;
    int triangleIndex = 0;

    // Generate vertices
    for (int i = 0; i <= steps; i++)
    {
        float theta = Mathf.PI * i / steps; 
        float y = radius * Mathf.Cos(theta); 
        float circleRadius = radius * Mathf.Sin(theta); 

        for (int j = 0; j <= pcount; j++)
        {
            float phi = 2 * Mathf.PI * j / pcount; 
            float x = circleRadius * Mathf.Cos(phi);
            float z = circleRadius * Mathf.Sin(phi);

            vertices[vertexIndex++] = new Vector3(x, y, z) + offset;
        }
    }

    // Generate triangles
    for (int i = 0; i < steps; i++)
    {
        for (int j = 0; j < pcount; j++)
        {
            int current = i * (pcount + 1) + j;
            int next = current + pcount + 1;

            triangles[triangleIndex++] = current;
            triangles[triangleIndex++] = next;
            triangles[triangleIndex++] = current + 1;

            triangles[triangleIndex++] = next;
            triangles[triangleIndex++] = next + 1;
            triangles[triangleIndex++] = current + 1;
        }
    }

    // Invert triangles if needed
    if (invert)
    {
        Array.Reverse(triangles);
    }

    return (vertices, triangles);
}
    public static (Vector3[], int[]) CreateSphere(float radius, bool invert = false)
        => CreateSphere(radius, Vector3.zero, invert);
    #endregion Implementation
}
