using System;
using UnityEngine;

public class TorusGenerator : ShapeGenerator
{
    #region InspectorVariables
    [Header("Torus values")]
    [Range(0, 100)] public float torusRadius = 1f;
    [Range(0, 10)] public float ringRadius = 0.1f;

    [Header("Technical")]
    [Range(8, 150)] public int ringSegmentCount = 50;
    [Range(3, 100)] public int segmentSideCount = 30;
    #endregion InspectorVariables

    #region Implementation
    protected override (Vector3[] vertices, int[] triangles) GenerateTopology()
        => CreateTorus(torusRadius, ringRadius, ringSegmentCount, segmentSideCount);

    private static (Vector3[] vertices, int[] triangles) CreateTorus(float torusRadius, float ringRadius, int ringSegmentCount, int segmentSideCount)
    {
        Vector3[] vertices = new Vector3[ringSegmentCount * segmentSideCount];
        int[] triangles = new int[ringSegmentCount * segmentSideCount * 6];

        float iStep = (2f * Mathf.PI) / ringSegmentCount;
        float jStep = (2f * Mathf.PI) / segmentSideCount;

        int t = 0;
        for (int i = 0; i < ringSegmentCount; i++)
            for (int j = 0; j < segmentSideCount; j++)
            {
                //generate vertices
                vertices[i * segmentSideCount + j] = GetPointOnTorus(torusRadius, ringRadius, i * iStep, j * jStep);

                //generate triangles
                triangles[t++] =
                    (i + 0) % ringSegmentCount * segmentSideCount
                    + (j + 0) % segmentSideCount;
                triangles[t++] =
                    (i + 1) % ringSegmentCount * segmentSideCount
                    + (j + 1) % segmentSideCount;
                triangles[t++] =
                    (i + 1) % ringSegmentCount * segmentSideCount
                    + (j + 0) % segmentSideCount;

                triangles[t++] =
                    (i + 0) % ringSegmentCount * segmentSideCount
                    + (j + 1) % segmentSideCount;
                triangles[t++] =
                    (i + 1) % ringSegmentCount * segmentSideCount
                    + (j + 1) % segmentSideCount;
                triangles[t++] =
                    (i + 0) % ringSegmentCount * segmentSideCount
                    + (j + 0) % segmentSideCount;
            }

        return (vertices, triangles);
    }

    private static Vector3 GetPointOnTorus(float torusRadius, float ringRadius, float u, float v)
    {
        float r = (torusRadius + ringRadius * Mathf.Sin(v));
        return new Vector3(r * Mathf.Sin(u), ringRadius * Mathf.Cos(v), r * Mathf.Cos(u));
    }
    #endregion Implementation
}