using System;
using System.Collections.Generic;

public static class IEnumerableExtensions
{
    public static IEnumerable<TAccumulate> PartialAggregate<TSource, TAccumulate>
    (
        this IEnumerable<TSource> source,
        TAccumulate seed,
        Func<TAccumulate, TSource, TAccumulate> func
    )
    {
        TAccumulate previous = seed;
        foreach (TSource item in source)
        {
            TAccumulate result = func(previous, item);
            yield return result;
            previous = result;
        }
    }

    public static IEnumerable<TAccumulate> FeedForwardUntil<TAccumulate>
    (
        TAccumulate seed,
        Func<TAccumulate, TAccumulate> func,
        Func<TAccumulate, bool> predicate
    )
    {
        TAccumulate previous = seed;
        while (!predicate(previous))
        {
            yield return previous;
            previous = func(previous);
        }

        yield return previous;
    }

    public static IEnumerable<TSource> WhereManual<TSource>
    (
        IEnumerable<TSource> source,
        Func<TSource, bool> predicate
    )
    {
        foreach (TSource item in source)
        {
            if (predicate(item))
                yield return item;
        }
    }

    public static IEnumerable<TSelect> SelectManual<TSource, TSelect>
    (
        IEnumerable<TSource> source,
        Func<TSource, TSelect> selector
    )
    {
        foreach (TSource item in source)
        {
            TSelect selected = selector(item);
            yield return selected;
        }
    }

    public static TAccumulate AggregateManual<TAccumulate, TSource>
    (
        IEnumerable<TSource> source,
        TAccumulate seed,
        Func<TAccumulate, TSource, TAccumulate> func
    )
    {
        TAccumulate previous = seed;
        foreach (TSource item in source)
        {
            previous = func(previous, item);
        }
        return previous;
    }

    public static IEnumerable<TSource> AppendRange<TSource>(this IEnumerable<TSource> source, IEnumerable<TSource> range)
    {
        IEnumerator<TSource> enum_source = source.GetEnumerator();
        while (enum_source.MoveNext())
            yield return enum_source.Current;
        enum_source.Dispose();

        if (range == null)
            yield break;

        IEnumerator<TSource> enum_range = range.GetEnumerator();
        while (enum_range.MoveNext())
            yield return enum_range.Current;
        enum_range.Dispose();
    }

    public static IEnumerable<TSource> AppendRangeAt<TSource>(this IEnumerable<TSource> source, IEnumerable<(uint index, TSource element)> indexed_range)
    {
        IEnumerator<TSource>
            enum_source = source.GetEnumerator();
        IEnumerator<(uint index, TSource element)>
            enum_range = indexed_range?.GetEnumerator();

        bool continue_source = enum_source.MoveNext();
        bool continue_range = enum_range != null && enum_range.MoveNext();

        for (int i = 0; continue_source || continue_range; i++)
        {
            if (continue_range
             && (!continue_source || i == enum_range.Current.index))
            {
                yield return enum_range.Current.element;
                continue_range = enum_range.MoveNext();
            }
            else if (continue_source)
            {
                yield return enum_source.Current;
                continue_source = enum_source.MoveNext();
            }
        }

        enum_source.Dispose();
        enum_range?.Dispose();
    }

    public static IEnumerable<TSource> SkipAt<TSource>(this IEnumerable<TSource> source, IEnumerable<int> range)
    {
        IEnumerator<TSource>
            enum_source = source.GetEnumerator();
        IEnumerator<int>
            enum_range = range?.GetEnumerator();

        bool continue_range = enum_range != null && enum_range.MoveNext();
        for (int i = 0; enum_source.MoveNext(); i++)
        {
            if (continue_range = continue_range
            && (enum_range.Current == i
             || (enum_range.MoveNext() && enum_range.Current == i)))
                continue;

            yield return enum_source.Current;
        }

        enum_source.Dispose();
        enum_range?.Dispose();
    }

    //// see https://stackoverflow.com/questions/1952153/what-is-the-best-way-to-find-all-combinations-of-items-in-an-array/10629938#10629938
    //public static IEnumerable<IEnumerable<T>> GetPermutationsWithRept<T>(this IEnumerable<T> source, int length)
    //{
    //    if (length == 1) 
    //        return source.Select(t => new T[] { t });

    //    return 
    //        GetPermutationsWithRept(source, length - 1)
    //        .SelectMany(t => source, (t1, t2) => t1.Concat(new T[] { t2 }));
    //}

    //public static IEnumerable<IEnumerable<T>> GetPermutations<T>(this IEnumerable<T> source, int length)
    //{
    //    if (length == 1) 
    //        return source.Select(t => new T[] { t });

    //    return 
    //        GetPermutations(source, length - 1)
    //        .SelectMany(
    //            t => source.Where(o => !t.Contains(o)),
    //            (t1, t2) => t1.Concat(new T[] { t2 }));
    //}

    //public static IEnumerable<IEnumerable<T>> GetKCombsWithRept<T>(this IEnumerable<T> source, int length) where T : IComparable
    //{
    //    if (length == 1) 
    //        return source.Select(t => new T[] { t });

    //    return 
    //        GetKCombsWithRept(source, length - 1)
    //        .SelectMany(
    //            t => source.Where(o => o.CompareTo(t.Last()) >= 0),
    //            (t1, t2) => t1.Concat(new T[] { t2 }));
    //}

    //public static IEnumerable<IEnumerable<T>> GetKCombs<T>(this IEnumerable<T> source, int length) where T : IComparable
    //{
    //    if (length == 1) 
    //        return source.Select(t => new T[] { t });

    //    return 
    //        GetKCombs(source, length - 1)
    //        .SelectMany(
    //            t => source.Where(o => o.CompareTo(t.Last()) > 0),
    //            (t1, t2) => t1.Concat(new T[] { t2 }));
    //}
}